// volume ops for for altered fdwt.c and idwt.c
// defined as static

#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include "RS_vol_ops.h"

int fdwtOutputSave = 0; //save the FDWT output data as FDWTtransformed.raw, FDWTtransformed.descr, FDWTtransformed.PGM
int fdwtCalcEntropy = 0; //save the FDWT output data memoryless entropy of all subbands in file FDWTtransformed.H0
int fdwtEntropyPredictor = 0; //number of predictor to apply before computing entropy 0-none, 1-(upper+left)/2, 2-MED
int idwtInputSave = 0; //save the IDWT input data as IDWTuntransformed.raw, IDWTuntransformed.descr, IDWTuntransformed.PGM


// for dwt_forward_transform and dwt_inverse_transform
// [level][subbandId] == 1 perform DWT part  (e.g., just prediction or just update) resulting in creation of given subband, == 0 skip this part, 
// if both parts of 1d DWT are skipped (e.g., 3 and 5, 7 and 11), reordering of coefficients is skipped as well in dwt_forward/inv_transform
// no subbandId==0, any val. is ok.
int tranfSubbandDo[MAX_LEVELS][15] = {
	// 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14 
	{ -0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -2, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -3, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
	{ -4, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1 },
};

const int transfOtherSubbandId[15] =				// for given subbandId contains subbandId of other step (0<->1) of DWT transformation 
// 0, 1, 2, 3, 4, 5, 6, 7, 8, 9,10,11,12,13,14 
{ -1, 2, 1, 5, 6, 3, 4,11,12,13,14, 7, 8, 9,10 };


// max transf. level at which denoising is performed, -1 for no denoising at all, levels are numbered from 0
// if level>maxLevel then denoising is called, but uses NOP method
int maxLevelDenoise = -1; 


// denoising arguments for levels and subbands, used by RS_vol_subband_denoise
// [level][subband][param], row elements: 0-subband id(must match row number), 1-denoising transform type tt, 
//														tt=0 - NOP, 2-6 ignore
//														tt=1 - soften 2-rx, 3-ry, 4-rz, 5-wc, 6-wn
//														tt=2 - median denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=3 - RCRS denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=4 - RCRS2 denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=5 - RCRS3 denoise 2-rx, 3-ry, 4-rz, 5-6 ignore
//														tt=6 - Mutekatsu Ryu denoising, 2-6 ignore
//														for details see RS_vol_subband_denoise()
//
//														tt=100-101 - skip the step, or both steps, see bottom of initialize_RS_vol_ops
int transfParams[MAX_LEVELS][15][7] = {
	// level 0
	   {{ -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1, 0, 2, 2, 1, 110, 1 },
		{ 2, 0, 2, 2, 1, 100, 1 },
		{ 3, 0, 1, 1, 1, 16, 1 }, //2d
		{ 4, 0, 2, 1, 2, 85, 1 },
		{ 5, 0, 1, 1, 1, 16, 1 }, //2d
		{ 6, 0, 2, 1, 2, 65, 1 },
		{ 7, 0, 1, 1, 1, 0, 0 }, //2d
		{ 8, 0, 1, 1, 1, 0, 0 }, //2d
		{ 9, 0, 1, 2, 2, 35, 1 },
		{ 10, 0, 1, 2, 2, 25, 1 },
		{ 11, 0, 1, 1, 1, 0, 0 }, //2d
		{ 12, 0, 1, 1, 1, 0, 0 }, //2d
		{ 13, 0, 1, 2, 2, 10, 1 },
		{ 14, 0, 1, 2, 2, 5, 1 },},
	// level 1
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1, 0, 2, 2, 1, 110, 1 },
		{ 2, 0, 2, 2, 1, 100, 1 },
		{ 3, 0, 1, 1, 1, 16, 1 }, //2d
		{ 4, 0, 2, 1, 2, 85, 1 },
		{ 5, 0, 1, 1, 1, 16, 1 }, //2d
		{ 6, 0, 2, 1, 2, 65, 1 },
		{ 7, 0, 1, 1, 1, 0, 0 }, //2d
		{ 8, 0, 1, 1, 1, 0, 0 }, //2d
		{ 9, 0, 1, 2, 2, 35, 1 },
		{ 10, 0, 1, 2, 2, 25, 1 },
		{ 11, 0, 1, 1, 1, 0, 0 }, //2d
		{ 12, 0, 1, 1, 1, 0, 0 }, //2d
		{ 13, 0, 1, 2, 2, 10, 1 },
		{ 14, 0, 1, 2, 2, 5, 1 }, },
	// level 2
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1, 0, 2, 2, 1, 110, 1 },
		{ 2, 0, 2, 2, 1, 100, 1 },
		{ 3, 0, 1, 1, 1, 16, 1 }, //2d
		{ 4, 0, 2, 1, 2, 85, 1 },
		{ 5, 0, 1, 1, 1, 16, 1 }, //2d
		{ 6, 0, 2, 1, 2, 65, 1 },
		{ 7, 0, 1, 1, 1, 0, 0 }, //2d
		{ 8, 0, 1, 1, 1, 0, 0 }, //2d
		{ 9, 0, 1, 2, 2, 35, 1 },
		{ 10, 0, 1, 2, 2, 25, 1 },
		{ 11, 0, 1, 1, 1, 0, 0 }, //2d
		{ 12, 0, 1, 1, 1, 0, 0 }, //2d
		{ 13, 0, 1, 2, 2, 10, 1 },
		{ 14, 0, 1, 2, 2, 5, 1 }, },
	// level 3
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1, 0, 2, 2, 1, 110, 1 },
		{ 2, 0, 2, 2, 1, 100, 1 },
		{ 3, 0, 1, 1, 1, 16, 1 }, //2d
		{ 4, 0, 2, 1, 2, 85, 1 },
		{ 5, 0, 1, 1, 1, 16, 1 }, //2d
		{ 6, 0, 2, 1, 2, 65, 1 },
		{ 7, 0, 1, 1, 1, 0, 0 }, //2d
		{ 8, 0, 1, 1, 1, 0, 0 }, //2d
		{ 9, 0, 1, 2, 2, 35, 1 },
		{ 10, 0, 1, 2, 2, 25, 1 },
		{ 11, 0, 1, 1, 1, 0, 0 }, //2d
		{ 12, 0, 1, 1, 1, 0, 0 }, //2d
		{ 13, 0, 1, 2, 2, 10, 1 },
		{ 14, 0, 1, 2, 2, 5, 1 }, },
	// level 4
		{ { -1, -1, -1, -1, -1, -1, -1 }, // no id==0
		{ 1, 0, 2, 2, 1, 110, 1 },
		{ 2, 0, 2, 2, 1, 100, 1 },
		{ 3, 0, 1, 1, 1, 16, 1 }, //2d
		{ 4, 0, 2, 1, 2, 85, 1 },
		{ 5, 0, 1, 1, 1, 16, 1 }, //2d
		{ 6, 0, 2, 1, 2, 65, 1 },
		{ 7, 0, 1, 1, 1, 0, 0 }, //2d
		{ 8, 0, 1, 1, 1, 0, 0 }, //2d
		{ 9, 0, 1, 2, 2, 35, 1 },
		{ 10, 0, 1, 2, 2, 25, 1 },
		{ 11, 0, 1, 1, 1, 0, 0 }, //2d
		{ 12, 0, 1, 1, 1, 0, 0 }, //2d
		{ 13, 0, 1, 2, 2, 10, 1 },
		{ 14, 0, 1, 2, 2, 5, 1 }, },
};





// initialize the module (e.g. load above data from file, check if level not to high), to be called at the beginning of dwt_forward_transform and dwt_inverse_transform
void initialize_RS_vol_ops(int maxLevel)
{
	FILE *f;
	int cmd=9999, level, subband, par, unfinished=1, read_err=0, nlevels;

	f = fopen(TRANSFPAR_DAT, "rt");

	if (!f)
		RS_err_exit("\nRS:initialize_RS_vol_ops:cannot open " TRANSFPAR_DAT);

	while (unfinished)
	{
		if (1 != fscanf(f, "%d", &cmd)) read_err = 1;
		switch (cmd)
		{

		case 1001:														// read maxLevelDenoise
			if (1 != fscanf(f, "%d", &maxLevelDenoise)) read_err = 1;
			break;

		case 1002:														// read fdwtOutputSave
			if (1 != fscanf(f, "%d", &fdwtOutputSave)) read_err = 1;
			break;

		case 1003:														// read idwtInputSave
			if (1 != fscanf(f, "%d", &idwtInputSave)) read_err = 1;
			break;

		case 1004:														// read fdwtCalcEntropy
			if (1 != fscanf(f, "%d", &fdwtCalcEntropy)) read_err = 1;
			break;

		case 1005:														// read fdwtEntropyPredictor
			if (1 != fscanf(f, "%d", &fdwtEntropyPredictor)) read_err = 1;
			break;

		case 1010:														// read transfParams
			for (level = 0; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					for (par = 0; par < 7; par++)
						if (1 != fscanf(f, "%d", &transfParams[level][subband][par])) read_err = 1;
			break;

		case 1011:														// read 1st page of transfParams, copy it to others
			for (subband = 0; subband < 15; subband++)
				for (par = 0; par < 7; par++)
					if (1 != fscanf(f, "%d", &transfParams[0][subband][par])) read_err = 1;
			for (level = 1; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					for (par = 0; par < 7; par++)
						transfParams[level][subband][par] = transfParams[0][subband][par];
			break;

		case 1012:														// read given number of transfParams pages
			if (1 != fscanf(f, "%d", &nlevels)) read_err = 1; 
			if (nlevels>MAX_LEVELS) { read_err = 1; break; }
			for (level = 0; level < nlevels; level++)
				for (subband = 0; subband < 15; subband++)
					for (par = 0; par < 7; par++)
						if (1 != fscanf(f, "%d", &transfParams[level][subband][par])) read_err = 1;
			break;

		case 1020:														// read tranfSubbandDo
			for (level = 0; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					if (1 != fscanf(f, "%d", &tranfSubbandDo[level][subband])) read_err = 1;
			break;

		case 1021:														// read 1st row of tranfSubbandDo, copy it to others
				for (subband = 0; subband < 15; subband++)
					if (1 != fscanf(f, "%d", &tranfSubbandDo[0][subband])) read_err = 1;
			for (level = 1; level < MAX_LEVELS; level++)
				for (subband = 0; subband < 15; subband++)
					tranfSubbandDo[level][subband] = tranfSubbandDo[0][subband];
			break;

		case 1022:														// read given number of tranfSubbandDo rows
			if (1 != fscanf(f, "%d", &nlevels)) read_err = 1;
			if (nlevels>MAX_LEVELS) { read_err = 1; break; }
			for (level = 0; level < nlevels; level++)
				for (subband = 0; subband < 15; subband++)
					if (1 != fscanf(f, "%d", &tranfSubbandDo[level][subband])) read_err = 1;
			break;

		case 9999:// all done
			unfinished = 0;
			break;

		default:
			RS_err_exit("\nRS:initialize_RS_vol_ops:bad command in " TRANSFPAR_DAT);
		}
		if (read_err)
			RS_err_exit("\nRS:initialize_RS_vol_ops:bad data in " TRANSFPAR_DAT);
	}

	if (maxLevel > MAX_LEVELS)												// partially check loaded data
		RS_err_exit("\nRS:initialize_RS_vol_ops:maxLevel too high.");

	if (fdwtOutputSave != 0 && fdwtOutputSave != 1)
		RS_err_exit("\nRS:initialize_RS_vol_ops: fdwtOutputSave != 0 && fdwtOutputSave != 1 ");
	if (idwtInputSave != 0 && idwtInputSave != 1)
		RS_err_exit("\nRS:initialize_RS_vol_ops: idwtInputSave != 0 && idwtInputSave != 1 ");

	for (level = 0; level < MAX_LEVELS; level++)
	{
		for (subband = 1; subband < 15; subband++)
		{
			if (tranfSubbandDo[level][subband] != 0 && tranfSubbandDo[level][subband] != 1)
				RS_err_exit("\nRS:initialize_RS_vol_ops: tranfSubbandDo[level][subband] != 0 && tranfSubbandDo[level][subband] != 1 ");
			if (transfParams[level][subband][0] != subband)
				RS_err_exit("\nRS:initialize_RS_vol_ops: transfParams[level][subband][0] != subband ");
		}
	}

	
	for (level = 0; level < MAX_LEVELS; level++)							// denoising modes 100 and 101 fed into transfSubbandDo
		for (subband = 0; subband < 15; subband++)
		{
			if (transfParams[level][subband][1] == 100 )		//set 0 in transfSubbandDo
				tranfSubbandDo[level][subband] = 0;
			else if (transfParams[level][subband][1] == 101)	//set 0 in transfSubbandDo for both DWT steps
			{
				const int pairSubband = transfOtherSubbandId[subband];
				tranfSubbandDo[level][subband] = 0;
				tranfSubbandDo[level][pairSubband] = 0;
			}
		}

	fclose(f);
}


// RS read volume pixel
t_sample RS_vol_get(const t_sample * dataBuffer, const int width, const int height, const int depth, const int x, const int y, const int z)
{
	assert(x >= 0 && x < width);
	assert(y >= 0 && y < height);
	assert(z >= 0 && z < depth);
	return dataBuffer[z*width*height + y*width + x];
}


// RS write volume pixel
void RS_vol_set(const t_sample sample, t_sample * dataBuffer, const int width, const int height, const int depth, const int x, const int y, const int z)
{
	assert(x >= 0 && x < width);
	assert(y >= 0 && y < height);
	assert(z >= 0 && z < depth);
	dataBuffer[z*width*height + y*width + x] = sample;
}


// RS copy pixels between vols, for even X start with even bX and use dX==2, for odd start with odd
void RS_copy_vol(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth,
	const int bx, const int ex, const int dx,	// band within data buffer - x coordinates range (from, to, step), from==to allowable but copies nothing
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz)	// z
{
	int x, y, z;

	assert(bx >= 0 && bx <= ex && ex <= width);
	assert(by >= 0 && by <= ey && ey <= height);
	assert(bz >= 0 && bz <= ez && ez <= depth);

	for (z = bz; z < ez; z += dz) /* slices */
		for (y = by; y < ey; y += dy) /* rows */
			for (x = bx; x < ex; x += dx) /* pixels */
			{
				unsigned bufOffset = z*width*height + y*width + x;
				toBuffer[bufOffset] = fromBuffer[bufOffset];
			}
}


// calculate weighted average of pixels distant from center(x,y,z) by rx, ry and rz, weighting center by wc, others by wo
int RS_vol_band_weighted_average(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz,	// neighbourhood size IN BAND COORINATES
	const int wc, const int wn				    // center and neighbours weights
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	long long int sum = 0;

	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					if (x == cx && y == cy && z == cz)
					{
						sum += wc*val.i;
						count += wc;
					}
					else
					{
						sum += wn*val.i;
						count += wn;
					}
				}

	assert(count>=0);
	if (sum<0)			// for rounding the average, not truncating
		sum -= count >> 1;
	else
		sum += count >> 1;
	return (int)(sum / count);
}


int __cmpint(const void *pi1, const void *pi2)
{
	return (*(int*)pi1)-(*(int*)pi2);
}


// calculate median of pixels distant from center(x,y,z) by rx, ry and rz
int RS_vol_band_median(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int median;
	int *tab = (int *)malloc(sizeof(int)*(rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));

	assert(tab);
	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
				}
	assert(count);

	qsort(tab, count, sizeof(int), __cmpint);
	median = tab[count >> 1];
	free(tab);
	return median;
}


// RCRS filter - return median if pixel value is outside of range of neighbours, original pixel otherwise
int RS_vol_band_RCRS(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int center = -99999;	// val of central point
	int median;
	int result;
	int *tab = (int *)malloc(sizeof(int)*(rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));


	assert(tab);
	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
					if (x == cx && y == cy && z == cz)
						center = val.i;

				}
	assert(count>1);

	qsort(tab, count, sizeof(int), __cmpint);
	median = tab[count >> 1];

	if (center == tab[0] && center != tab[1] || center == tab[count - 1] && center != tab[count - 2])
		result = median;
	else
		result = center;

	free(tab);
	return result;
}



// RCRS 2 filter - ret. closest neighbor if pixel value is outside of range of neighbours, original pixel otherwise
int RS_vol_band_RCRS2(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int center = -99999;	// val of central point
	int result;
	int *tab = (int *)malloc(sizeof(int)*(rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));


	assert(tab);
	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
					if (x == cx && y == cy && z == cz)
						center = val.i;

				}
	assert(count>1);

	qsort(tab, count, sizeof(int), __cmpint);

	if (center == tab[0])
		result = tab[1];
	else if (center == tab[count - 1])
		result = tab[count - 2];
	else
		result = center;

	free(tab);
	return result;
}


// RCRS 3 filter - ret. closest neighbor different then pixel value, but not crossing the median of neibourhood
int RS_vol_band_RCRS3(const t_sample * dataBuffer, const int width, const int height, const int depth, // data buffer
	const int bx, const int ex, const int dx,	// band within data buffer - allowable x coordinates range (from, to, step)
	const int by, const int ey, const int dy,	// y
	const int bz, const int ez, const int dz,	// z
	const int cx, const int cy, const int cz,	// reference voxel position in dataBuffer coordinates
	const int rx, const int ry, const int rz	// neighbourhood size IN BAND COORINATES
	)
{
	int x, y, z;	// neighbour  x, y and z, 
	t_sample val;	// placeholder for value,
	int count = 0;
	int center = -99999;	// val of central point
	int result;
	int *tab = (int *)malloc(sizeof(int)*(rx * 2 + 1)*(ry * 2 + 1)*(rz * 2 + 1));


	assert(tab);
	for (z = cz - dz*rz; z <= cz + dz*rz; z += dz)
		for (y = cy - dy*ry; y <= cy + dy*ry; y += dy)
			for (x = cx - dx*rx; x <= cx + dx*rx; x += dx)
				if (x >= bx && x < ex && y >= by && y < ey && z >= bz && z < ez)
				{
					val = RS_vol_get(dataBuffer, width, height, depth, x, y, z);
					tab[count++] = val.i;
					if (x == cx && y == cy && z == cz)
						center = val.i;

				}
	assert(count>=4);

	qsort(tab, count, sizeof(int), __cmpint);

	if (center == tab[0])
	{
		int i=1;
		const int fstop = (count & 0x01) ? (count / 2) : (count / 2) - 1 ;

		while (i < fstop && tab[i] == tab[0]) // idac od poczatku znajdz el. inny niz pierwszy, ale nie powyzej polowy wszystkich
			i++;

		result = tab[i];
	}
	else if (center == tab[count - 1])
	{
		int i = count - 2;
		const int bstop = count / 2;

		while (i > bstop && tab[i] == tab[count - 1]) // idac od konca znajdz pierwszy inny niz ostatni, ale nie ponizej polowy wszystkich
			i--;

		result = tab[i];
	}
	else
		result = center;

	free(tab);
	return result;
}



//// RS dump vol to stdout
//void RS_dump_vol(char* descr, const t_sample * fromBuffer, const int width, const int height, const int depth)
//{
//	int x, y, z;
//
//	printf("\nvol: %s, width: %d, height: %d, depth: %d", descr, width, height, depth);
//
//	for (z = 0; z < depth; z++) /* slices */
//	{
//		for (y = 0; y < height; y++) /* rows */
//		{
//			printf("\n");
//			for (x = 0; x < width; x++) /* pixels */
//			{
//				printf(" % 3d", RS_vol_get(fromBuffer, width, height, depth, x, y, z));
//			}
//		}
//		printf("\n");
//	}
//}






// save subband, optionally wih description in another file and optionally save first slice as 8-bit pgm 
void RS_save_subband(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int subbandId,
	const char *volName,
	const int SaveDescription, const char *descrName,
	const int SavePGM, const char * PGMName,
	const int reversible)									
{
	int min, max;
	int x, y, z;
	FILE *f;

	if (reversible)
		min = max = RS_vol_get(fromBuffer, width, height, depth, sx, sy, sz).i;
	else
		min = max = (int) RS_vol_get(fromBuffer, width, height, depth, sx, sy, sz).f;

	f = fopen(volName, "wb");										// save volume as raw
	for (z = sz; z < ez; z += dz) /* For all slices */				// calc min/max and save samples
		for (y = sy; y < ey; y += dy) /* For all rows */
			for (x = sx; x < ex; x += dx) /* For all columns */
			{
				int i;
				t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z);
				if (reversible)
					i = s.i;
				else
					i = (int) s.f;
				if(i < min)
					min = i;
				if (i>max)
					max = i;
				if (1 != fwrite(&s, sizeof(s), 1, f))
					RS_err_exit("\nRS:RS_save_subband problem writing raw file.");
			}
	fclose(f);

	if (SaveDescription)
	{
		f = fopen(descrName, "wt");
		fprintf(f,
			"VOLUME (tile) \nwidth: %d, height: %d, depth: %d,\n\n"
			"TRANSF \nlevel: %d, subbandId: %d,\n"
			"fwd: %d, reversible: %d, \n\n"
			"SUBBAND in volume\n"
			"bx: %d, ex: %d, dx: %d, sx: %d,\n"
			"by: %d, ey: %d, dy: %d, sy: %d,\n"
			"bz: %d, ez: %d, dz: %d, sz: %d,\n"
			"width: %d, height: %d, depth: %d,\n"
			"volName: %s, \n\n"
			"SAMPLE (as int) \nmin: %d, max: %d. \n\n",
			width, height, depth,
			level, subbandId,
			fwd, reversible,
			bx, ex, dx, sx,
			by, ey, dy, sy,
			bz, ez, dz, sz,
			(dx - 1 + ex - sx) / dx, (dy - 1 + ey - sy) / dy, (dz - 1 + ez - sz) / dz,
			volName,
			min, max);
		fclose(f);
	}

	if (SavePGM)
	{
		f = fopen(PGMName, "wb");										// save first volume slice as 8-bit PGM
		fprintf(f, "P5%c%d %d%c%d%c", 0x0A, (dx - 1 + ex - sx) / dx, (dy - 1 + ey - sy) / dy, 0x0A, 255, 0x0A);//header
		//for (z = sz; z < ez; z += dz) /* For all slices */	commented out - just first xy slice!
		z = sz;
			for (y = sy; y < ey; y += dy) /* For all rows */
				for (x = sx; x < ex; x += dx) /* For all columns */
				{
					int i;
					unsigned char uc;
					t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z);
					if (reversible)
						i = s.i;
					else
						i = (int)s.f;

					if (min != max)
						uc = (unsigned char)(((256 * (long long)(i - min))) / (max - min));
					else
						uc = 0;
					if (1 != fwrite(&uc, 1, 1, f))
						RS_err_exit("\nRS:RS_save_subband problem writing PGM file.");
				}
		fclose(f);
	}
}
 

// denoises given subband of band of 3d transformation part
void RS_vol_subband_denoise(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int subbandId, 	// caled from FDWT? (orherwise IDWT) (inform.), transform level and subband Id (DENOISING step) (1-14)
	const int reversible)									// will reversible DWT be called?
{
	const int 
		tt = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][1],  //if level>maxLevelDenoise then it will copy from fromBuffer to toBuffer
		rx = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][2],
		ry = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][3],
		rz = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][4],
		wc = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][5],
		wn = level>maxLevelDenoise ? 0 : transfParams[level][subbandId][6];

	int x, y, z;

	assert(maxLevelDenoise < MAX_LEVELS);
	assert(level < MAX_LEVELS);
	assert(subbandId >= 1 && subbandId <= 14);
	assert(transfParams[level][subbandId][0] == subbandId);


	if (tt == 100 ||	// w initialize_RS_vol_ops przepisywane na 0 do transfSubbandDo, a wiec nie bedzie DWT, wiec wynik odszumienia bedzie nadpisany nieodszumionym
		tt == 101)	    // w initialize_RS_vol_ops przepisywane na 0 do transfSubbandDo dla obu czci (step 0<->1) DWT 
	{
		return;
	}


	if (tt != 0 && !reversible)
		RS_err_exit("\nRS:RS_vol_subband_denoise denoising only for int transforms .");

	//obeys maxLevelDenoise
	for (z = sz; z < ez; z+=dz) /* For all slices */												
		for (y = sy; y < ey; y += dy) /* For all rows */
			for (x = sx; x < ex; x+=dx) /* For all columns */
			{
				t_sample s = RS_vol_get(fromBuffer, width, height, depth, x, y, z);

				if (tt == 0)
				{
				}
				else if (tt == 1)
				{
					s.i = RS_vol_band_weighted_average(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz,
						wc, wn);
				}
				else if (tt == 2)
				{
					s.i = RS_vol_band_median(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 3)
				{
					s.i = RS_vol_band_RCRS(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 4)
				{
					s.i = RS_vol_band_RCRS2(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 5)
				{
					s.i = RS_vol_band_RCRS3(fromBuffer, width, height, depth,  //read from dataCopy
						bx, ex, dx, by, ey, dy, bz, ez, dz,
						x, y, z,
						rx, ry, rz);
				}
				else if (tt == 6)
				{
					s.i = 0;
				}
				else
					RS_err_exit("\nRS:RS_vol_subband_denoise bad transform type.");

				RS_vol_set(s, toBuffer, width, height, depth, x, y, z);
			}
}




// RS ret. number of L coefficients at the beginning of the non-sparse coefficients range from u0 to u1 (without u1)
int RS_calc_middle(const unsigned int u0, const unsigned int u1)
{
	const unsigned int boundary = u1 - u0;
	return ((u0 % 2) == 0) ? ((boundary + 1) >> 1) : (boundary >> 1);

}


// finds subbandId for given transPart and 1d DWT coordinates
int RS_transPart_to_subbandId(
	const int by, const int ey, // band y boundaries
	const int bz, const int ez, // z
	const int level, const int transPart, 	// transform level and 3D TRANSFORM PART (1-6)
	const int y, const int z, // 3d coordinates y and z of any voxel to be transformed using 1d DWT
	const int decompositionsZ, const int decompositionsY)	// number of decompositions for Z and Y (mind order of them)
{
	int fHz, fHy; // index of the first H sample in z and y dir. == ez/ey if no decomposition in given direction

	assert(transPart >= 1 && transPart <= 6);

	if (transPart == 1 || transPart == 2)
		return transPart;

	if (level < decompositionsZ)  // podpasma dla Z
		fHz = bz + RS_calc_middle(bz, ez);
	else
		fHz = ez;

	if (transPart == 3 || transPart == 4) // kierunek y
	{
		const int subbandBaseId = (transPart == 3 ? 3 : 5);

		if (z >= bz && z<fHz)
			return subbandBaseId;

		if (z >= fHz && z<ez)
			return subbandBaseId + 1;

		assert(0);
	}

	if (level < decompositionsY)  // podpasma dla Y
		fHy = by + RS_calc_middle(by, ey);
	else
		fHy = ey;

	if (transPart == 5 || transPart == 6) // kierunek x
	{
		const int subbandBaseId = (transPart == 5 ? 7 : 11);

		if (z >= bz && z<fHz && y >= by && y<fHy)
			return subbandBaseId;

		if (z >= bz && z<fHz && y >= fHy && y<ey)
			return subbandBaseId + 1;

		if (z >= fHz && z<ez && y >= by && y<fHy)
			return subbandBaseId + 2;

		if (z >= fHz && z<ez && y >= fHy && y<ey)
			return subbandBaseId + 3;

		assert(0);
	}
	assert(0);
	return -1;
}



// denoises band for given 3d transformation part, due to earlier transformations at given level band may have subbands 
void RS_vol_band_denoise_TRANSPART(const t_sample * fromBuffer, t_sample * toBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex, const int dx, const int sx,	// band within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int level, const int transPart, 	// caled from FDWT? (orherwise IDWT) (inform.), transform level and 3D TRANSFORM PART (1-6)
	const int decompositionsZ, const int decompositionsY, 	// number of decompositions for Z and Y
	const int reversible)									// will reversible DWT be called?
{
	int fHz, fHy; // index of the first H sample in z and y dir. == ez/ey if no decomposition in given direction

	assert(sx == bx || sx == bx + 1);
	assert(sy == by || sy == by + 1);
	assert(sz == bz || sz == bz + 1);
	assert(transPart >= 1 && transPart <= 6);

	if (transPart == 1 || transPart == 2)
	{
		    RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, bz, ez, dz, sz, fwd, level, transPart, reversible);
			assert(transPart   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , sz   , decompositionsZ, decompositionsY));
			assert(transPart   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , ez-1 , decompositionsZ, decompositionsY));
		return;
	}
	
	assert(dz == 1);	// skok o 2 moze byc tylko dla X lub Y
	assert(bz == sz);
	if (level < decompositionsZ)  // podpasma dla Z
		fHz = bz + RS_calc_middle(bz, ez);
	else
		fHz = ez;

	if (transPart == 3 || transPart == 4) // kierunek y
	{
		int subbandId = (transPart == 3 ? 3 : 5);

  		    RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, bz, fHz, dz, sz, fwd, level, subbandId, reversible);
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , sz   , decompositionsZ, decompositionsY));
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , fHz-1, decompositionsZ, decompositionsY));

		if (fHz!=ez)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, ey, dy, sy, fHz, ez, dz, fHz, fwd, level, subbandId + 1, reversible);
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , fHz  , decompositionsZ, decompositionsY));
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , ez-1 , decompositionsZ, decompositionsY));
		}

		return;
	}

	assert(dy == 1);	// skok o 2 moze byc tylko dla X 
	assert(by == sy);
	if (level < decompositionsY)  // podpasma dla Y
		fHy = by + RS_calc_middle(by, ey);
	else
		fHy = ey;

	if (transPart == 5 || transPart == 6) // kierunek x
	{
		int subbandId = (transPart == 5 ? 7 : 11);

		    RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, fHy, dy, sy, bz, fHz, dz, sz, fwd, level, subbandId, reversible);
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , sz   , decompositionsZ, decompositionsY));
			assert(subbandId   == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy-1, fHz-1, decompositionsZ, decompositionsY));

		if (fHy != ey)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, fHy, ey, dy, fHy, bz, fHz, dz, sz, fwd, level, subbandId + 1, reversible);
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy  , sz   , decompositionsZ, decompositionsY));
			assert(subbandId+1 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 , fHz-1, decompositionsZ, decompositionsY));
		}
		if (fHz != ez)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, by, fHy, dy, sy, fHz, ez, dz, fHz, fwd, level, subbandId + 2, reversible);
			assert(subbandId+2 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, sy   , fHz  , decompositionsZ, decompositionsY));
			assert(subbandId+2 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy-1, ez-1 , decompositionsZ, decompositionsY));
		}
		if (fHz != ez && fHy != ey)
		{
			RS_vol_subband_denoise(fromBuffer, toBuffer, width, height, depth, bx, ex, dx, sx, fHy, ey, dy, fHy, fHz, ez, dz, fHz, fwd, level, subbandId + 3, reversible);
			assert(subbandId+3 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, fHy  ,  fHz  , decompositionsZ, decompositionsY));
			assert(subbandId+3 == RS_transPart_to_subbandId(by, ey, bz, ez, level, transPart, ey-1 ,  ez-1 , decompositionsZ, decompositionsY));
		}

		return;
	}
}

void RS_err_exit(const char * s)  // output error msg and exit
{
	fflush(stdout);
	fflush(stderr);
	fputs(s, stderr);
	fflush(stderr);
	exit(-1);
}


// calculate entropy of a subband being an image (3d not implemented), only 3x5 reversible
double RS_calc_subband_entropy(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
	int bx, int ex,
	int by, int ey,
	const int predictor)
{
#define TABSIZE 0x10000
	int tab[TABSIZE];
	int x, y, i, sample, predicted, index;

	double invlog2 = 1.0 / log(2.0);  // hepler 
	const int nSamples = (ex-bx) * (ey-by); // helper
	double H0 = 0.0;//result

	for (i = 0; i < TABSIZE; i++)
		tab[i] = 0;

	for (y= by; y< ey; y++)
		for (x = bx; x < ex; x++)
		{
			sample = RS_vol_get(fromBuffer, width, height, depth, x, y, 0).i;
			predicted = 0;

			if (predictor != 0)
			{
				if ((x > bx) && (y > by))
					switch (predictor)
					{
					case 1: predicted = ( RS_vol_get(fromBuffer, width, height, depth, x, y - 1, 0).i + 
						                  RS_vol_get(fromBuffer, width, height, depth, x - 1, y, 0).i )/2;
							break;
					case 2:
							{
								const int
									a = RS_vol_get(fromBuffer, width, height, depth, x - 1, y, 0).i,
									b = RS_vol_get(fromBuffer, width, height, depth, x, y - 1, 0).i,
									c = RS_vol_get(fromBuffer, width, height, depth, x - 1, y - 1, 0).i;
								int min_ab, max_ab;

								if (a < b)
								{
									min_ab = a;
									max_ab = b;
								}
								else
								{
									min_ab = b;
									max_ab = a;
								}

								if (c >= max_ab)
									predicted = min_ab;
								else if (c <= min_ab)
									predicted = max_ab;
								else
									predicted = a + b - c;
							}
						break;
					default: RS_err_exit("RS_calc_subband_entropy() bad predictor");
					}
				else if ((x == bx) && (y>by))
					predicted = RS_vol_get(fromBuffer, width, height, depth, x, y-1, 0).i;  // upper
				else if ((x > bx) && (y == by))
					predicted = RS_vol_get(fromBuffer, width, height, depth, x - 1, y, 0).i;  // left-hand
				else
					predicted = 0;
			}


			index = (TABSIZE >> 1) + sample - predicted;
			if (index<0 || index >= TABSIZE)
				RS_err_exit("RS_calc_subband_entropy() bad min/max");
			tab[index]++;
		}

	for (i = 0; i < TABSIZE; i++)
		if (tab[i])
			H0 -= (double)(tab[i]) * log((double)(tab[i]) / (double)(nSamples)) * invlog2;
	return H0;
}


//#define ENTROPY_TIMING

// calculate entropy of transformed volume being an image (3d not implemented), only 53x5 reversible, as sum of entropies of subbands, 
// using prediction errors of samples, and save the result to outFileName
void RS_calc_entropy(const t_sample * fromBuffer, const int width, const int height, const int depth, // data buffers
	const int bx, const int ex_in, const int dx, const int sx,	// subband within data buffer - allowable x coordinates range (from, to, step, starting position)
	const int by, const int ey_in, const int dy, const int sy,	// y
	const int bz, const int ez, const int dz, const int sz,	// z
	const int fwd, const int maxLevel, const int reversible, const int predictor, const char *outFileName) // maxLevel is number of levels
{
	int ns; // number of summed subbands
	double H0;	//H0 cumulative
	int lvl;
	FILE *f;
	int ex, ey; 


#ifdef ENTROPY_TIMING
	clock_t start = clock();
#endif	

	ex = ex_in;
	ey = ey_in;
	ns = 0;
	H0 = 0.0;

	if (!(bx == 0 && dx == 1 && sx == 0)
		|| !(by == 0 && dy == 1 && sy == 0)
		|| !(bz == 0 && dz == 1 && sz == 0 && ez == 1)
		|| !(depth == 1)
		|| !(fwd)
		|| !(reversible)
		) RS_err_exit("RS_calc_entropy() bad call arguments");

	for (lvl = 0; lvl < maxLevel; lvl++)
	{
		int fHx = RS_calc_middle(bx, ex);
		int fHy = RS_calc_middle(by, ey);

		H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, fHx, ex, by, fHy, predictor); // HL
		H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, bx, fHx, fHy, ey, predictor); // LH
		H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, fHx, ex, fHy, ey, predictor); // HH
		ns += 3;

		ex = fHx;
		ey = fHy;
	}
	H0 += RS_calc_subband_entropy(fromBuffer, width, height, depth, bx, ex, by, ey, predictor); // LL
	ns++;

#ifdef ENTROPY_TIMING
	printf("\nEntropy calculated in %f seconds\n", (float)(clock() - start) / CLOCKS_PER_SEC);
#endif

	f = fopen(outFileName, "wt");
	fprintf(f, "%.0f (H0, bits), predictor: %d subbands: %d\n\n", ceil(H0), predictor, ns);
	fclose(f);
}

